--
-- 3. Propositions and Proofs
--

-- Propositions as Types

-- https://stackoverflow.com/questions/69182724/lean-4-unknown-identifier-proof
-- The code below actually does not work
-- #check And
-- #check Or
-- #check Not
-- #check Implies
--
-- variable (p q r : Prop)
-- #check And p q
-- #check Or p q
-- #print Or
-- #check Implies (And p q) (And q p)
-- #check p → q

-- Working with Propositions as Types
variable {p : Prop}
variable {q : Prop}

-- theorem t1 : p → q → p := fun hp : p => fun hq : q => hp
theorem t1 : p → q → p :=
  fun hp : p =>
  fun hq : q =>
  show p from hp

#print t1

variable (p q r s : Prop)

#check t1 p q
#check t1 r s

-- Propositional Logic

-- conj
variable (p q : Prop)

example (hp : p) (hq : q) : p ∧ q := And.intro hp hq

#check fun (hp : p) (hq : q) => And.intro hp hq

variable (p q : Prop)

example (h : p ∧ q) : p := And.left h
example (h : p ∧ q) : q := And.right h

example (h : p ∧ q) : q ∧ p :=
  And.intro (And.right h) (And.left h)

variable (hp : p) (hq : q)
#check (⟨hp, hq⟩ : p ∧ q)

example (h : p ∧ q) : q ∧ p :=
  ⟨h.right, h.left⟩

example (h : p ∧ q) : q ∧ p ∧ q :=
  ⟨h.right, ⟨h.left, h.right⟩⟩

example (h : p ∧ q) : q ∧ p ∧ q :=
  ⟨h.right, h.left, h.right⟩

-- disjunction
variable (p q : Prop)
example (hp : p) : p ∨ q := Or.intro_left q hp
example (hq : q) : p ∨ q := Or.intro_right p hq

example (h : p ∨ q) : q ∨ p :=
  Or.elim h
    (fun hp : p =>
      show q ∨ p from Or.intro_right _ hp)
    (fun hq : q =>
      show q ∨ p from Or.intro_left _ hq)

example (h : p ∨ q) : q ∨ p :=
  Or.elim h (fun hq => Or.inr hq) (fun hp => Or.inl hp)

example (h : p ∨ q) : q ∨ p :=
  h.elim (fun hq => Or.inr hq) (fun hp => Or.inl hp)

-- Negation and Falsity
variable (p q : Prop)

example (hpq : p → q) (hnq : ¬q) : ¬p :=
  fun hp : p =>
  show False from hnq (hpq hp)

example (hp : p) (hnp : ¬p) : q := False.elim (hnp hp)

example (hp : p) (hnp : ¬p) : q := absurd hp hnp

example (hnp : ¬p) (hq : q) (hqp : q → p) : r :=
  absurd (hqp hq) hnp

-- Logical Equivalence
theorem and_swap : p ∧ q ↔ q ∧ p :=
  Iff.intro
    (fun h : p ∧ q =>
     show q ∧ p from And.intro h.right h.left)
    (fun h : q ∧ p =>
     show p ∧ q from ⟨h.right, h.left⟩)

example : p ∧ q ↔ q ∧ p :=
  ⟨fun h => ⟨h.right, h.left⟩, fun h => ⟨h.right, h.left⟩⟩

#check and_swap p q
#check and_swap q p

variable (h : p ∧ q)
example : q ∧ p := (and_swap p q).mp h

-- Introducing Auxiliary Subgoals
example (h : p ∧ q) : q ∧ p :=
  have hp : p := h.left
  have hq : q := h.right
  show q ∧ p from And.intro hq hp

example (h : p ∧ q) : q ∧ p :=
  have hp : p := h.left
  suffices hq : q from And.intro hq hp
  show q from And.right h

-- Classical Logic
open Classical in
#check em p

open Classical in
theorem dne {p : Prop} (h : ¬¬p) : p :=
  Or.elim (em p)
    (fun hp : p => hp)
    (fun hnp : ¬p => absurd hnp h)

open Classical in
example (h : ¬¬p) : p :=
  byCases
    (fun h1 : p => h1)
    (fun h2 : ¬p => absurd h2 h)

open Classical in
example (h : ¬¬p) : p :=
  byContradiction
    (fun h1 : ¬p =>
     show False from h h1)

open Classical in
example (h : ¬(p ∧ q)) : ¬p ∨ ¬q :=
  Or.elim (em p)
    (fun hp : p =>
     Or.inr
       (show ¬q from
         fun hq : q =>
         h ⟨hp, hq⟩))
    (fun hp : ¬p =>
     Or.inl hp)

-- Exercises
-- ∧ と ∨ の可換性
example : p ∧ q ↔ q ∧ p :=
  Iff.intro
    (fun h : p ∧ q =>
      ⟨h.right, h.left⟩)
    (fun h : q ∧ p =>
      ⟨h.right, h.left⟩)
example : p ∨ q ↔ q ∨ p :=
  Iff.intro
    (fun h : p ∨ q =>
      h.elim
        (fun h : p => Or.inr h)
        (fun h : q => Or.inl h))
    (fun h : q ∨ p =>
      h.elim
        (fun h : q => Or.inr h)
        (fun h : p => Or.inl h))

-- ∧ と ∨ の結合性
example : (p ∧ q) ∧ r ↔ p ∧ (q ∧ r) :=
  have h1 : (p ∧ q) ∧ r → p ∧ (q ∧ r) :=
    (fun h =>
      ⟨h.left.left, ⟨h.left.right, h.right⟩⟩)
  have h2 : p ∧ (q ∧ r) → (p ∧ q) ∧ r :=
    (fun h =>
      ⟨⟨h.left, h.right.left⟩, h.right.right⟩)
  ⟨h1, h2⟩
example : (p ∨ q) ∨ r ↔ p ∨ (q ∨ r) :=
  have h1 : (p ∨ q) ∨ r → p ∨ (q ∨ r) :=
    (fun h =>
      h.elim
        (fun hpq =>
          hpq.elim
            (fun hp => Or.inl hp)
            (fun hq => Or.inr (Or.inl hq)))
        (fun hr => Or.inr (Or.inr hr)))
  have h2 : p ∨ (q ∨ r) → (p ∨ q) ∨ r :=
    (fun h =>
      h.elim
        (fun hp =>
          Or.inl (Or.inl hp))
        (fun hqr =>
          hqr.elim
            (fun hq => Or.inl (Or.inr hq))
            (fun hr => Or.inr hr)))
  Iff.intro h1 h2

-- 分配性
example : p ∧ (q ∨ r) ↔ (p ∧ q) ∨ (p ∧ r) :=
  have h1 : p ∧ (q ∨ r) → (p ∧ q) ∨ (p ∧ r) :=
    (fun h =>
      h.right.elim
        (fun hq => Or.inl (And.intro h.left hq))
        (fun hr => Or.inr (And.intro h.left hr)))
  have h2 : (p ∧ q) ∨ (p ∧ r) → p ∧ (q ∨ r) :=
    (fun h =>
      h.elim
        (fun hpq => And.intro hpq.left (Or.inl hpq.right))
        (fun hpr => And.intro hpr.left (Or.inr hpr.right)))
  Iff.intro h1 h2
example : p ∨ (q ∧ r) ↔ (p ∨ q) ∧ (p ∨ r) :=
  have h1 :=
    (fun h =>
      h.elim
        (fun h => ⟨Or.inl h, Or.inl h⟩)
        (fun h => ⟨Or.inr h.left, Or.inr h.right⟩))
  have h2 :=
    (fun h =>
      have h2l : p ∨ q := h.left
      have h2r : p ∨ r := h.right
      h2l.elim
        (fun hp => Or.inl hp)
        (fun hq =>
          h2r.elim
            (fun hp => Or.inl hp)
            (fun hr => Or.inr (And.intro hq hr))))
  Iff.intro h1 h2

-- 他の性質
example : (p → (q → r)) ↔ (p ∧ q → r) :=
  have h1 : (p → q → r) → p ∧ q → r :=
    (fun h =>
      fun hpq =>
        (h hpq.left) hpq.right)
  have h2 : (p ∧ q → r) → p → q → r :=
    (fun h =>
      fun hp =>
        fun hq =>
          h ⟨hp, hq⟩)
  Iff.intro h1 h2
example : ((p ∨ q) → r) ↔ (p → r) ∧ (q → r) :=
  have h1 : (p ∨ q → r) → (p → r) ∧ (q → r) :=
    fun h =>
      have hpr :=
        (fun hp =>
          h (Or.inl hp))
      have hqr :=
        (fun hq =>
          h (Or.inr hq))
      And.intro hpr hqr
  have h2 : (p → r) ∧ (q → r) → p ∨ q → r :=
    fun h =>
      fun h' : p ∨ q =>
        h'.elim
          (fun hp => h.left hp)
          (fun hq => h.right hq)
  Iff.intro h1 h2
example : ¬(p ∨ q) ↔ ¬p ∧ ¬q :=
  have h1 : ¬(p ∨ q) → ¬p ∧ ¬q :=
    fun h =>
      have hnp : ¬p :=
        fun hp =>
          show False from h (Or.inl hp)
      have hnq : ¬q :=
        fun hq =>
          show False from h (Or.inr hq)
      And.intro hnp hnq
  have h2 : ¬p ∧ ¬q → ¬(p ∨ q) := 
    fun h =>
      fun hpq : p ∨ q =>
        show False from
          hpq.elim
            (fun hp => h.left hp)
            (fun hq => h.right hq)
  Iff.intro h1 h2
example : ¬p ∨ ¬q → ¬(p ∧ q) :=
  fun h =>
    h.elim
      (fun hnp =>
        fun hpq =>
          show False from hnp (hpq.left))
      (fun hnq => sorry)
example : ¬(p ∧ ¬p) := sorry
example : p ∧ ¬q → ¬(p → q) := sorry
example : ¬p → (p → q) := sorry
example : (¬p ∨ q) → (p → q) := sorry
example : p ∨ False ↔ p := sorry
example : p ∧ False ↔ False := sorry
example : (p → q) → (¬q → ¬p) := sorry

open Classical in
example : (p → q ∨ r) → ((p → q) ∨ (p → r)) :=
  fun h =>
    (em q).elim
      (fun hq =>
        Or.inl (fun hp => hq))
      (fun hnq =>
        have h1 : p → r :=
          fun hp : p =>
            (h hp).elim
              (fun hq =>
                absurd hq hnq)
              (fun hr => hr)
        Or.inr h1)
open Classical in
example : ¬(p ∧ q) → ¬p ∨ ¬q :=
  fun hnpq =>
    sorry
open Classical in
example : ¬(p → q) → p ∧ ¬q := sorry
open Classical in
example : (p → q) → (¬p ∨ q) := sorry
open Classical in
example : (¬q → ¬p) → (p → q) := sorry
open Classical in
example : p ∨ ¬p := sorry
open Classical in
example : (((p → q) → p) → p) := sorry
